import base64
import hmac
import hashlib
import os
import re
from typing import List, Union
from warnings import warn

import smartcar.config as config
import smartcar.helpers as helpers
import smartcar.types as types

API_VERSION = "2.0"


def set_api_version(version: str) -> None:
    """
    Update the version of Smartcar API you are using

    Args:
        version (str): the version of the api you want to use
    """
    if re.match(r"\d+\.\d+", version):
        global API_VERSION
        API_VERSION = version
    else:
        raise ValueError(
            rf"Version '{version}' must match regex '\d+\.\d+' .  e.g. '2.0', '1.0'"
        )


def get_api_version() -> str:
    """
    Returns:
        Version of API requests are currently being sent to.

        NOTE: May not reflect the version of Smartcar API that instantiated smartcar.Vehicle's
        are sending requests to.
    """

    global API_VERSION
    return API_VERSION


def get_user(access_token: str) -> types.User:
    """
    Retrieve the userId associated with the access_token

    Args:
        access_token (str): Smartcar access token

    Returns:
        User: NamedTuple("User", [("id", str), ("meta", namedtuple)])

    Raises:
        SmartcarException
    """
    url = f"{config.API_URL}/v{API_VERSION}/user"
    headers = {"Authorization": f"Bearer {access_token}"}
    response = helpers.requester("GET", url, headers=headers)

    return types.select_named_tuple("user", response)


def get_vehicles(access_token: str, paging: dict = None) -> types.Vehicles:
    """
    Get a list of the user's vehicle ids

    Args:
        access_token (str): A valid access token from a previously retrieved
            access object

        paging (dictionary, optional): Can include "limit" and "offset" keys:
            limit (int, optional): The number of vehicle ids to return

            offset (int, optional): The index to start the vehicle list at

    Returns:
        Vehicles: NamedTuple("Vehicles", [("vehicles", List[str]), ("paging", Paging), ("meta", namedtuple)])

    Raises:
        SmartcarException
    """
    url = f"{config.API_URL}/v{API_VERSION}/vehicles"
    headers = {"Authorization": f"Bearer {access_token}"}
    params = paging if paging is not None else None
    response = helpers.requester("GET", url, headers=headers, params=params)

    return types.select_named_tuple("vehicles", response)


def get_compatibility(
    vin: str, scope: List[str], country: str = "US", options: dict = None
) -> Union[types.CompatibilityV1, types.CompatibilityV2]:
    """
    Verify if a vehicle (vin) is eligible to use Smartcar. Use to confirm whether
    specific vehicle is compatible with the permissions provided.

    A compatible vehicle is one that:
        1. Has hardware required for internet connectivity
        2. Belongs to the makes and models Smartcar is compatible with
        3. Is compatible with the required permissions (scope) that your app is requesting
            access to

    Note: The `mode` and `test_mode_compatibility_level` options arguments are only valid for Smartcar API v1.0
            and `test_mode` has been deprecated

    Args:
        vin (str)

        scope (List[str]): List of scopes (permissions) -> to check if vehicle is compatible

        country (str, optional)

        options (dict): Can contain client_id, client_secret, and flags.
            client_id (str, optional)

            client_secret (str, optional)

            version (str): Version of API you want to use

            flags (dict - {str: bool}, optional): An optional list of feature flags

            test_mode (bool, optional): Deprecated, please use `mode` instead.
                Launch Smartcar Connect in [test mode](https://smartcar.com/docs/guides/testing/).

            mode (str, optional): Determine what mode Smartcar Connect should be launched in.
                Should be one of test, live or simulated.

            test_mode_compatibility_level (str, optional): This parameter is required when the API is invoked in test mode.
                Possible values with details are documented in our Integration Guide.

    Returns:
        CompatibilityV1: NamedTuple("Compatibility", [("compatible", bool), ("meta", namedtuple)])
        OR
        CompatibilityV2: NamedTuple("Compatibility",
            [
                ("compatible", bool),
                ("reason", Union[str, None]),
                ("capabilities", List[dict]),
                ("meta", namedtuple),
            ],
        )
    """
    client_id = os.environ.get("SMARTCAR_CLIENT_ID")
    client_secret = os.environ.get("SMARTCAR_CLIENT_SECRET")
    api_version = API_VERSION
    params = {"vin": vin, "scope": " ".join(scope), "country": country}

    # Configuring options.
    if options is None:
        helpers.validate_env()
    else:
        # client_id and client_secret passed in options dict() will take precedence
        # over environment variables.
        client_id = options.get("client_id", client_id)
        client_secret = options.get("client_secret", client_secret)
        api_version = options.get("version", api_version)

        if options.get("flags"):
            flags_str = helpers.format_flag_query(options["flags"])
            params["flags"] = flags_str

        if options.get("version"):
            api_version = options["version"]

        if api_version == "1.0":
            if options.get("test_mode") is not None:
                warn(
                    'The "testMode" parameter is deprecated, please use the "mode" parameter instead.',
                    DeprecationWarning,
                )
                params["mode"] = "test" if options.get("test_mode") else "live"
            elif options.get("mode"):
                params["mode"] = options.get["mode"]

            if options.get("test_mode_compatibility_level"):
                params["test_mode_compatibility_level"] = options[
                    "test_mode_compatibility_level"
                ]
                params["mode"] = "test"

            if params.mode not in ["test", "live", "simulated"]:
                raise Exception(
                    "The \"mode\" parameter MUST be one of the following: 'test', 'live', 'simulated'",
                )

    # Ensuring client_id and client_secret are present
    if client_id is None or client_secret is None:
        raise Exception(
            "'get_compatibility' requires a client_id AND client_secret. "
            "Either set these as environment variables, OR pass them in as part of the 'options'"
            "dictionary. The recommended course of action is to set up environment variables"
            "with your client credentials. i.e.: "
            "'SMARTCAR_CLIENT_ID' and 'SMARTCAR_CLIENT_SECRET'"
        )

    url = f"{config.API_URL}/v{api_version}/compatibility"

    # Configuring for compatibility endpoint
    id_secret = f"{client_id}:{client_secret}"
    encoded_id_secret = id_secret.encode("ascii")
    base64_bytes = base64.b64encode(encoded_id_secret)
    base64_id_secret = base64_bytes.decode("ascii")
    headers = {"Authorization": f"Basic {base64_id_secret}"}

    response = helpers.requester("GET", url, headers=headers, params=params)

    if api_version == "1.0":
        return types.select_named_tuple("compatibility_v1", response)
    elif api_version == "2.0":
        return types.select_named_tuple("compatibility_v2", response)
    else:
        raise Exception("Please use a valid API version (e.g. '1.0' or '2.0')")


# ===========================================
# Webhook functions
# ===========================================


def hash_challenge(amt: str, challenge: str) -> str:
    """
    Take in a randomly generated challenge string, and use an
    Application Management Token as a key to be hashed.

    Args:
        amt (str): Application Management Token from Smartcar Dashboard

        challenge: Randomly generated string from smartcar after requesting
            a challenge.

    Returns:
        hex-encoding of resulting hash
    """
    h = hmac.new(amt.encode(), challenge.encode(), hashlib.sha256)
    return h.hexdigest()


def verify_payload(amt: str, signature: str, body: str) -> bool:
    """
    Verify webhook payload against AMT and signature

    Args:
        amt (str): Application Management Token from Smartcar Dashboard

        signature: sc-signature header value

        body: Stringified JSON of the webhook response body

    Returns:
        Boolean
    """
    return hash_challenge(amt, body) == signature


# ===========================================
# Management
# ===========================================
def get_management_token(amt: str, username: str = "default") -> str:
    secret = f"{username}:{amt}"
    encoded_secret = secret.encode("ascii")
    base64_bytes = base64.b64encode(encoded_secret)
    return base64_bytes.decode("ascii")


def get_connections(
    amt: str, filter: dict = {}, paging: dict = {}
) -> types.GetConnections:
    """
    Returns a paged list of all the vehicles that are connected to the application
    associated with the management API token used sorted in descending order by connection date.

    Args:
        amt (str): Application Management Token from Smartcar Dashboard

        filter (dict, optional)
            vehicle_id (str, optional)
            user_id (str, optional)

        paging (dict, optional)
            limit (int, optional)
            cursor_id (str, optional)

    Returns:
        GetConnections = NamedTuple("GetConnections", [
            ("connections", List[Connection]),
            ("paging", PagingCursor),
            ("meta", namedtuple)
            ],
        )
    """
    params = {}
    if filter.get("user_id"):
        params["user_id"] = filter.get("user_id")
    if filter.get("vehicle_id"):
        params["vehicle_id"] = filter.get("vehicle_id")
    if paging.get("cursor"):
        params["cursor"] = filter.get("cursor")
    if paging.get("limit"):
        params["limit"] = filter.get("limit")

    url = f"{config.MANAGEMENT_API_URL}/v{get_api_version()}/management/connections/"
    headers = {"Authorization": f"Basic {get_management_token(amt)}"}
    response = helpers.requester("GET", url, headers=headers, params=params)
    data = response.json()
    connections = [
        types.Connection(c.get("vehicleId"), c.get("userId"), c.get("connectedAt"))
        for c in data["connections"]
    ]

    response_paging = data.get("paging", {})
    response_paging = types.PagingCursor(response_paging.get("cursor"))

    return types.GetConnections(
        connections,
        response_paging,
        types.build_meta(response.headers),
    )


def delete_connections(amt: str, filter: dict = {}) -> types.DeleteConnections:
    """
    Deletes all the connections by vehicle or user ID and returns a list
    of all connections that were deleted.

    Args:
        amt (str): Application Management Token from Smartcar Dashboard

        filter (dict, optional): Can contain EITHER vehicle_id OR user_id
            vehicle_id (str, optional)
            user_id (str, optional)

    Returns:
        DeleteConnections = NamedTuple("DeleteConnections", [
            ("connections", List[Connection]),
            ("meta", namedtuple)
            ],
        )

    """
    user_id = filter.get("user_id")
    vehicle_id = filter.get("vehicle_id")
    if user_id and vehicle_id:
        raise Exception("Filter can contain EITHER user_id OR vehicle_id, not both")

    params = {}
    if user_id:
        params["user_id"] = user_id
    elif vehicle_id:
        params["vehicle_id"] = vehicle_id

    url = f"{config.MANAGEMENT_API_URL}/v{get_api_version()}/management/connections/"
    headers = {"Authorization": f"Basic {get_management_token(amt)}"}
    response = helpers.requester("DELETE", url, headers=headers, params=params)
    data = response.json()
    connections = [
        types.Connection(c.get("vehicleId"), c.get("userId"), c.get("connectedAt"))
        for c in data["connections"]
    ]

    return types.DeleteConnections(
        connections,
        types.build_meta(response.headers),
    )



import datetime
from collections import namedtuple
from typing import List, NamedTuple, Union
import re
import requests.structures as rs
import enum

# Return types for Smartcar API.
#
# 'generate_named_tuple' is used to generate an un-typed namedtuple from
# a dictionary. It will return a namedtuple that has attributes matching
# the dictionary's keys. Use this function as a catch all, or for data
# that does not have an explicit length (e.g. response headers, batch requests)
#
# Otherwise, use the explicitly defined NamedTuples for better type hints!


def generate_named_tuple(
    dictionary: dict, name: str = "namedtuple", kebab_case=False
) -> namedtuple:
    """
    Take a dictionary and map its keys to the attributes of a named tuple.

    Args:
        dictionary (dict): Any dictionary
        name (str): The desired name of the returned namedtuple
        kebab_case (bool): format kebab-cased keys, otherwise handle camelCased keys

    Returns:
        namedtuple: With attributes matching the keys of the inputted dictionary
    """

    # Instantiate keys list, which will keep order consistent
    keys = dictionary.keys()
    if len(keys) == 0:
        return None

    attributes = []

    for key in keys:
        # Convert kebab to snake case, if relevant
        if kebab_case:
            formatted = key.replace("-", "_").lower()

        # convert camel to snake case (using regex function)
        else:
            formatted = _camel_to_snake(key)

        attributes.append(formatted)

    gen = namedtuple(name, attributes)

    return gen._make([dictionary[k] for k in keys])


def _camel_to_snake(camel_string: str) -> str:
    """
    Use regex to change camelCased string to snake_case

    Args:
        camel_string(str)

    Returns:
        A snake_cased string

    """
    result = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", camel_string)
    result = re.sub("(.)([0-9]+)", r"\1_\2", result)
    result = re.sub("([a-z0-9])([A-Z])", r"\1_\2", result)
    return result.lower()


def build_meta(response_headers: rs.CaseInsensitiveDict) -> namedtuple:
    smartcar_headers = {
        "sc-data-age": "data_age",
        "sc-unit-system": "unit_system",
        "sc-request-id": "request_id",
    }

    meta_dict = {}
    for key, value in smartcar_headers.items():
        if key in response_headers:
            meta_dict[value] = response_headers[key]

    return generate_named_tuple(meta_dict, "Meta", True)


# ===========================================
# auth_client.py
# ===========================================

Access = NamedTuple(
    "Access",
    [
        ("access_token", str),
        ("token_type", str),
        ("expires_in", int),
        ("expiration", datetime.datetime),
        ("refresh_token", str),
        ("refresh_expiration", datetime.datetime),
    ],
)


def make_access_object(access: dict) -> Access:
    return Access(
        access.get("access_token"),
        access.get("token_type"),
        access.get("expires_in"),
        access.get("expiration"),
        access.get("refresh_token"),
        access.get("refresh_expiration"),
    )


# ===========================================
# smartcar.py
# ===========================================

Paging = NamedTuple("Paging", [("count", int), ("offset", int)])

PagingCursor = NamedTuple("PagingCursor", [("cursor", Union[int, None])])

User = NamedTuple("User", [("id", str), ("meta", namedtuple)])

Vehicles = NamedTuple(
    "Vehicles",
    [("vehicles", List[str]), ("paging", Paging), ("meta", namedtuple)],
)


class Reason(enum.Enum):
    VEHICLE_NOT_COMPATIBLE = "VEHICLE_NOT_COMPATIBLE"
    MAKE_NOT_COMPATIBLE = "MAKE_NOT_COMPATIBLE"


Capability = NamedTuple(
    "Capability",
    [
        ("permission", str),
        ("endpoint", str),
        ("capable", bool),
        ("reason", Union[Reason, None]),
    ],
)


def format_capabilities(capabilities_list: List[dict]) -> List[Capability]:
    typed_capabilities_map = map(
        lambda capability: Capability(
            capability["permission"],
            capability["endpoint"],
            capability["capable"],
            capability["reason"],
        ),
        capabilities_list,
    )  # Returns iterable map
    typed_capabilities_list = list(typed_capabilities_map)
    return typed_capabilities_list


CompatibilityV1 = NamedTuple(
    "Compatibility", [("compatible", bool), ("meta", namedtuple)]
)

CompatibilityV2 = NamedTuple(
    "Compatibility",
    [
        ("compatible", bool),
        ("reason", Union[Reason, None]),
        ("capabilities", List[Capability]),
        ("meta", namedtuple),
    ],
)

# ===========================================
# vehicle.py
# ===========================================

Vin = NamedTuple("Vin", [("vin", str), ("meta", namedtuple)])

Charge = NamedTuple(
    "Charge",
    [("is_plugged_in", bool), ("state", str), ("meta", namedtuple)],
)

ChargeLimit = NamedTuple("ChargeLimit", [("limit", float), ("meta", namedtuple)])

Battery = NamedTuple(
    "Battery",
    [("percent_remaining", float), ("range", float), ("meta", namedtuple)],
)

BatteryCapacity = NamedTuple(
    "BatteryCapacity", [("capacity", float), ("meta", namedtuple)]
)

Fuel = NamedTuple(
    "Fuel",
    [
        ("range", float),
        ("percent_remaining", float),
        ("amount_remaining", float),
        ("meta", namedtuple),
    ],
)

TirePressure = NamedTuple(
    "TirePressure",
    [
        ("front_left", int),
        ("front_right", int),
        ("back_left", int),
        ("back_right", int),
        ("meta", namedtuple),
    ],
)

EngineOil = NamedTuple("EngineOil", [("life_remaining", float), ("meta", namedtuple)])

Odometer = NamedTuple("Odometer", [("distance", float), ("meta", namedtuple)])

Location = NamedTuple(
    "Location",
    [("latitude", float), ("longitude", float), ("meta", namedtuple)],
)

Attributes = NamedTuple(
    "Attributes",
    [
        ("id", str),
        ("make", str),
        ("model", str),
        ("year", str),
        ("meta", namedtuple),
    ],
)

Action = NamedTuple("Action", [("status", str), ("message", str), ("meta", namedtuple)])

Status = NamedTuple("Status", [("status", str), ("meta", namedtuple)])

Permissions = NamedTuple(
    "Permissions", [("permissions", list), ("paging", Paging), ("meta", namedtuple)]
)

Subscribe = NamedTuple(
    "Subscribe",
    [("webhook_id", str), ("vehicle_id", str), ("meta", namedtuple)],
)

Connection = NamedTuple(
    "Connection",
    [("vehicle_id", str), ("user_id", str), ("connected_at", Union[str, None])],
)

GetConnections = NamedTuple(
    "GetConnections",
    [("connections", List[Connection]), ("paging", PagingCursor), ("meta", namedtuple)],
)

DeleteConnections = NamedTuple(
    "DeleteConnections",
    [("connections", List[Connection]), ("meta", namedtuple)],
)

Response = NamedTuple("Response", [("body", dict), ("meta", namedtuple)])

# ===========================================
# Lock Status Tuples
# ===========================================

Door = NamedTuple("Door", [("type", str), ("status", str)])

Window = NamedTuple("Window", [("type", str), ("status", str)])

Sunroof = NamedTuple("Sunroof", [("type", str), ("status", str)])

Storage = NamedTuple("Storage", [("type", str), ("status", str)])

ChargingPort = NamedTuple("ChargingPort", [("type", str), ("status", str)])

LockStatus = NamedTuple(
    "LockStatus",
    [
        ("is_locked", bool),
        ("doors", List[Door]),
        ("windows", List[Window]),
        ("sunroof", List[Sunroof]),
        ("storage", List[Storage]),
        ("charging_port", List[ChargingPort]),
        ("meta", namedtuple),
    ],
)

# ===========================================
# Named Tuple Selector Function
# ===========================================


def select_named_tuple(path: str, response_or_dict) -> NamedTuple:
    """
    This function is used to select one of the pre-defined NamedTuples
    based on a path provided. Upon selection, the appropriate NamedTuple
    will be instantiated and returned. This function can take in a
    response from Smartcar API OR a dictionary with "body", "headers",
    "path", and "status".

    The only use case for the parsing of a dictionary (rather than a response)
    would be for "batch" calls to Smartcar API. Upon sending a batch request
    to SmartcarAPI, a single response is returned. The response data
    (i.e. response.json()) contains a list of dictionaries, each dictionary
    representing the result of each request in the batch. For this reason,
    this function needs to be able to parse a dictionary as well.

    Note that if a path is not dictated in one of the
    conditionals below, the raw data will be returned. This, in theory,
    shouldn't run because paths are defined by the contributing developer.
    In the case of "batch" requests, incorrect paths to batch will result in
    a SmartcarException before this function is called.

    Args:
        path (str): Smartcar API path
        response_or_dict: Smartcar response, or a dictionary after parsing
            the response to the "batch" endpoint

    Returns:
        NamedTuple: Appropriate to the path.

    """

    if type(response_or_dict) == dict:
        headers_dict = rs.CaseInsensitiveDict(response_or_dict["headers"])
        headers = build_meta(headers_dict)
        data = response_or_dict["body"]
    else:
        headers = build_meta(response_or_dict.headers)
        data = response_or_dict.json()

    # smartcar.py
    if path == "user":
        return User(data["id"], headers)

    elif path == "vehicles":
        return Vehicles(
            data["vehicles"],
            Paging(data["paging"]["count"], data["paging"]["offset"]),
            headers,
        )

    elif path == "compatibility_v1":
        return CompatibilityV1(data["compatible"], headers)

    elif path == "compatibility_v2":
        typed_capabilities = format_capabilities(data["capabilities"])
        return CompatibilityV2(
            data["compatible"], data["reason"], typed_capabilities, headers
        )

    # vehicle.py
    elif path == "vin":
        return Vin(data["vin"], headers)

    elif path == "charge":
        return Charge(data["isPluggedIn"], data["state"], headers)

    elif path == "battery":
        return Battery(data["percentRemaining"], data["range"], headers)

    elif path == "battery/capacity":
        return BatteryCapacity(data["capacity"], headers)

    elif path == "fuel":
        return Fuel(
            data["range"],
            data["percentRemaining"],
            data["amountRemaining"],
            headers,
        )

    elif path == "tires/pressure":
        return TirePressure(
            data["frontLeft"],
            data["frontRight"],
            data["backLeft"],
            data["backRight"],
            headers,
        )

    elif path == "engine/oil":
        return EngineOil(data["lifeRemaining"], headers)

    elif path == "odometer":
        return Odometer(data["distance"], headers)

    elif path == "location":
        return Location(data["latitude"], data["longitude"], headers)

    elif path == "charge/limit":
        return ChargeLimit(data["limit"], headers)

    elif path == "permissions":
        return Permissions(
            data["permissions"],
            Paging(data["paging"]["count"], data["paging"]["offset"]),
            headers,
        )
    elif path == "security":
        return LockStatus(
            data["isLocked"],
            data["doors"],
            data["windows"],
            data["sunroof"],
            data["storage"],
            data["chargingPort"],
            headers,
        )
    elif path == "subscribe":
        return Subscribe(data["webhookId"], data["vehicleId"], headers)

    elif (
        path == "lock"
        or path == "unlock"
        or path == "start_charge"
        or path == "stop_charge"
        or path == "set_charge_limit"
        or path == "send_destination"
    ):
        return Action(data["status"], data["message"], headers)

    elif path == "disconnect" or path == "unsubscribe":
        return Status(data["status"], headers)

    elif path == "":
        return Attributes(
            data["id"],
            data["make"],
            data["model"],
            data["year"],
            headers,
        )

    elif path == "request":
        return Response(data, headers)

    elif type(data) == dict:
        return generate_named_tuple(data, "Data")

    else:
        return data



import os

API_URL = os.environ.get("SMARTCAR_API_ORIGIN", "https://api.smartcar.com")
AUTH_URL = os.environ.get(
    "SMARTCAR_AUTH_ORIGIN", "https://auth.smartcar.com/oauth/token"
)
CONNECT_URL = os.environ.get("SMARTCAR_CONNECT_ORIGIN", "https://connect.smartcar.com")
MANAGEMENT_API_URL = os.environ.get(
    "SMARTCAR_MANAGEMENT_API_ORIGIN", "https://management.smartcar.com"
)



import os
import platform
import requests

import smartcar.exception as sce
from smartcar import __version__


def requester(method: str, url: str, **kwargs) -> requests.models.Response:
    """
    Attaches the kwargs into the headers, sends the request to the Smartcar API
        and handles all error cases

    Args:
        method (str): HTTP method

        url (str): url of the request

        **kwargs: parameters for the request

    Returns:
        requests.models.Response: response from the request to the Smartcar API
    """
    if "headers" not in kwargs:
        kwargs["headers"] = {}

    kwargs["headers"]["User-Agent"] = (
        f"Smartcar/{__version__}({platform.system()}; "
        f"{platform.machine()}) Python v{platform.python_version()}"
    )

    try:
        response = requests.request(method, url, timeout=310, **kwargs)
        code = response.status_code
        headers = response.headers
        body = response.text

        if response.ok:
            return response
        else:
            raise sce.exception_factory(code, headers, body)

    except Exception as e:
        if isinstance(e, sce.SmartcarException):
            raise e
        else:
            raise sce.SmartcarException(message="SDK_ERROR") from e


def validate_env(mode: str = "live") -> None:
    """
    Helper Function to determine if environment variables for client id
    and secret are set properly.

    Args:
        mode: str ['test'|'simulated'|'live']

    Raises:
        Basic Exception
    """
    prefix = "E2E_SMARTCAR" if mode != "live" else "SMARTCAR"

    if (
        f"{prefix}_CLIENT_ID" not in os.environ
        or f"{prefix}_CLIENT_SECRET" not in os.environ
    ):
        raise Exception(
            f'"{prefix}_CLIENT_ID" and "{prefix}_CLIENT_SECRET"'
            f" environment variables must be set"
        )


def format_flag_query(flags: dict) -> str:
    """
    Takes a dictionary of flags and parses it into a space separated
    string:  "<key>:<value> <key>:<value> ..." to be injected
    as a query parameter.

    Args:
        flags: dict

    Returns:
        str, space separated with "<key>:<value"

        e.g.
        flags == {"flag1 : True, "color" : "green"}
        -> "flag1:True color:green:
    """
    flags_str = ""

    for flag_name, flag_value in flags.items():
        if type(flag_value) == bool:
            flag_value = str(flag_value).lower()
        flags_str += f"{flag_name}:{flag_value} "

    return flags_str.strip()


def format_path_and_attribute_for_batch(raw_path: str) -> tuple:
    """
    Prettify the batch attribute and path names.
    Returned formatted_path will have the slash sliced off.
    Returned formatted_attribute will be attached to the final return of vehicle.batch().
    The naming of the attribute should consider empty and nested paths.

    Args:
        raw_path: Raw path (minus the slash) to smartcar endpoint

    Returns:
        (<formatted path>, <formatted attribute>)

        e.g.
        1. "EMPTY" raw_path  == '/' -> ('', 'attributes')
        2. "NORMAL" raw_path == '/odometer' -> ('odometer', 'odometer')
        3. "NESTED" raw_path == '/engine/oil' -> ('engine/oil', 'engine_oil')
    """
    # mapper holds unique situations where the path does not exactly line up with the function to call
    # we have a set_charge_limit but are not concerned with it in batch calls
    mapper = {
        "charge/limit": "get_charge_limit",
        "tires/pressure": "tire_pressure",
        "": "attributes",
        "security": "lock_status",
    }
    formatted_path = raw_path[1:] if raw_path[0] == "/" else raw_path
    formatted_attribute = mapper.get(formatted_path, formatted_path)
    if "/" in formatted_attribute:
        formatted_attribute = formatted_attribute.replace("/", "_")

    return formatted_path, formatted_attribute
